#define __STDIO_FDEVOPEN_COMPAT_12


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "util_macros.h"
#include "crypto.h"
#include "ledcomm.h"
#include "comm.h"

#define COMM_EN		1
#if COMM_EN
#else
	#define printf(...)
#endif

/*
uint32_t ee_key[4] EEMEM = { 0x6b8b4567L, 0x327b23c6L, 0xf3aaa47L, 0x2c613bbfL };
uint32_t key[4];
*/
uint32_t wait_period;

#if COMM_EN
int put_byte(char c) {
	if(c == '\n') {
		put_byte('\r');
	}
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
	return 0;
}

int get_byte() {
	return 0;
}
#endif

#if COMM_EN
void key_init_uart() 
{
  UBRRL = 103;    // Set the baud rate
//  UBRRL = 52;
	//DDRB = 0xff;

	set_bit(UCSRB, TXEN);
	fdevopen( put_byte, get_byte, 0 );
}
#endif

void key_initialize() {
#if COMM_EN
	key_init_uart();
#endif


#ifndef __AVR_ATtiny2313__
	//set_bit(GICR, INT1);
	set_bit(GICR, INT1);

	unset_bit(MCUCR, ISC10);
	unset_bit(MCUCR, ISC11);

	unset_bit(MCUCR, SM0);
	set_bit(MCUCR, SM1);
	unset_bit(MCUCR, SM2);
#else
	set_bit(EIFR, INTF1);
//	set_bit(EIFR, INTF1);

#endif

	/* Setup timer0 to actually call into LEDComm */
	// Get timer0 working
	TCNT0 = 0;          // Clear the counter register

#ifndef __AVR_ATtiny2313__
	set_bit(TIMSK, OCIE0);
	set_bit(TCCR0, WGM01); // CTC mode
	set_bit(TCCR0, CS01);  // prescale: CLK_IO / 8
	OCR0 = 198;
#else
	set_bit(TIMSK, OCIE0A);

	// ctc mode
	set_bit(TCCR0A, WGM01);
	set_bit(TCCR0B, CS01);

	OCR0A = 99;
#endif 
//	eeprom_read_block((void *) key, (void *) &ee_key, sizeof(key));


	ledcomm_init();

	sei();
}

#ifndef __AVR_ATtiny2313__
#define TIMER0_SIGNAL_WE_WANT TIMER0_COMP_vect
#else
#define TIMER0_SIGNAL_WE_WANT TIMER0A_COMP_vect
#endif

SIGNAL(TIMER0_SIGNAL_WE_WANT)
{
  //DDRB = 0xff;
  //PORTB = 1;
	ledcomm_tick();
}

/* bcr6: Do nothing on either of the external interrupts,
 * but make sure that we wake up when we get them */
SIGNAL(INT1_vect) {
#if 0
	if(bit_is_set(PORTLED, LEDTX)) {
		unset_bit(PORTLED, LEDTX);
	} else {
		set_bit(PORTLED, LEDTX);

	}
#endif

}

#define MAX_WP 2000000
enum {
	STATE_SLEEP = 1,
	STATE_WAITLOCK = 2,
	STATE_WAITTRANS = 4,
	STATE_RESPOND = 8
};

int main() {
	uint32_t key[] = { 0x6b8b4567L, 0x327b23c6L, 0xf3aaa47L, 0x2c613bbfL };
	uint16_t id = 13456;
	uint32_t nonce[2];
	uint32_t wait_period;
	char s;

//DDRLED = 0xff;
DDRD = 0x00;
	unset_bit(PORTLED, LEDRX);

	key_initialize();	


	printf("LEDKey Starting ....\n");
	s = STATE_SLEEP;
	while(1)
	{
		wait_period = 0;
		switch(s) {
			case STATE_SLEEP:
				printf("going into sleep mode\n");
			  	ledcomm_sleep();

				// put the device to sleep until someone 
				// actually hits the button.
				set_bit(MCUCR, SE);
				sleep_mode();
				ledcomm_unsleep(); 
				ledcomm_init();
				s = STATE_WAITLOCK;
				printf("wc\n");
			case STATE_WAITLOCK:
				// when we wake up, wait to clock sync or time out
				while(!ledcomm_client_present()) {
					wait_period++;
					if(wait_period >= MAX_WP) 
					{
					//	printf("wctos\n");
						s = STATE_SLEEP;
						break;
					}
				}

				if(s & 1) {
					break;
				}

				printf("gtwd\n");
				s = STATE_WAITTRANS;

			case STATE_WAITTRANS:
				// wait for a nonce or timeout
				while(ledcomm_rx_len() < LEN_CMD_NONCE) {
					wait_period++;
					if(wait_period >= MAX_WP) 
					{
					//	printf("wdtos\n");
						s = STATE_SLEEP;
						break;
					}
				}

				if(s & 1) break;

				printf("got %d\n", LEN_CMD_NONCE);
				s = STATE_RESPOND;	

			case STATE_RESPOND:
			{
				volatile char *tmp = ledcomm_get_rxbuf();
				uint32_t tnonce[2], enonce[2];
				char tmp2[LEN_CMD_NONCE];

				// respond appropriately to the challenge
				memcpy((char * ) tnonce, (char *) tmp+1, sizeof(nonce));

				tea_decipher_32(tnonce, nonce, key);

				nonce[0] += 1;
				nonce[1] += id;

				tea_encipher_32(nonce, enonce, key);

				tmp2[0] = CMD_CRYPT;
				memcpy((char *) tmp2+1, &id, 2);	
				memcpy((char *) tmp2+3, enonce, sizeof(enonce));

				printf(">> resp (%d)\n", tmp2[0]);
				printf("nonce=%lx,%lx\n", nonce[0], nonce[1]);

				ledcomm_rx_reset();
				int III = ledcomm_send_buffer(tmp2, LEN_CMD_CRYPT);
				printf("send_buffer() ret %d\n", III);
				s = STATE_SLEEP;
			}
		}

	}
}
