#define __STDIO_FDEVOPEN_COMPAT_12

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "util_macros.h"
#include "random.h"
#include "crypto.h"
#include "ledcomm.h"
#include "wiegand.h"
#include "comm.h"

#define STATUSDDR      DDRC
#define STATUSPORT     PORTC
#define STATUSREDPIN   2
#define STATUSGREENPIN 0
#define STATUSBLUEPIN  1
#define STATUSRED      4
#define STATUSBLUE     2
#define STATUSGREEN    1

#define WIEGANDDDR     DDRC


volatile static uint8_t rd_state;
volatile static uint8_t rd_timeout;

#define 	UB_LEN	254
static char uart_buf[UB_LEN];
static uint8_t uart_pos;
static uint8_t uart_len;
volatile static uint8_t cs_count;
volatile static uint8_t ms_count;

enum {
	STATE_WAITING,
	STATE_NONCE_SENT
};


int get_byte() {
	return 0;
}


void send_byte() {
	if(uart_len == 0) {
		unset_bit(UCSRB, UDRIE);
	} else {
		char cb = uart_buf[uart_pos];
		int newpos = ((int) uart_pos + 1) % UB_LEN;
		uart_pos = newpos;

		if(uart_len > 0) {
			uart_len--;
		}

		set_bit(UCSRB, UDRIE);
		//PORTB = ~uart_len;
		UDR = cb;
	}
	return;
}


int put_byte(char c) {
	if(c == '\n') {
		put_byte('\r');
	}

	// bcr6: The buffer is full, busy loop
	while(UB_LEN == uart_len) {
	}

	unset_bit(UCSRB, UDRIE);
	int comb = ((int) uart_pos + (int) uart_len);
	int pos = comb % UB_LEN;

	//PORTB = ~pos;
	uart_buf[pos] = c;
	uart_len++;
	set_bit(UCSRB, UDRIE);

	return 0;
}


#if 0
int put_byte(char c) {
	if(c == '\n') {
		put_byte('\r');
	}
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
	return 0;
}
#endif

SIGNAL(USART_UDRE_vect) {
//PORTB = ~cs_count;
	send_byte();
	//UDR = 'x';
}

SIGNAL(TIMER0_COMP_vect)
{

	ledcomm_tick();
	ms_count++;
	if(ms_count >= 100) {
		ms_count = 0;
		if(rd_timeout != 0) {
			cs_count++;
			if(cs_count >= rd_timeout) {
				printf("TIMING OUT!!!\n");
				rd_state = STATE_WAITING;
				cs_count = 0;
				rd_timeout = 0;
				ledcomm_rx_reset();
			}

		}
	}
}

void reader_init_uart() 
{
	uart_pos = 0;
	uart_len = 0;

	UBRRL = 103;    // Set the baud rate
	DDRB = 0xff;

	set_bit(UCSRB, TXEN);
	fdevopen( put_byte, get_byte, 0 );
}

void reader_initialize() {
	reader_init_uart();
	//PORTB = 0xf0;
	rg_init();

	cs_count = 0;
	ms_count = 0;


	/* Setup timer0 to actually call into LEDComm */
	// Get timer0 working
	TCNT0 = 0;          // Clear the counter register
	set_bit(TIMSK, OCIE0);
	set_bit(TCCR0, WGM01); // CTC mode
	set_bit(TCCR0, CS01);  // prescale: CLK_IO / 8
	OCR0 = 198;

	ledcomm_init();
	rd_state = STATE_WAITING;
	sei();
}

/* Change state, and stay there for ms_timeout milliseconds, 
 * or return to the waiting state
 *
 */
void change_state(char state, uint8_t ms_timeout) {
	rd_state = state;		
	rd_timeout = ms_timeout;
	cs_count = 0;
}

uint8_t get_state() {
	return rd_state;
}

int main() {
	reader_initialize();	
	//uint32_t x[] = { 3122222UL, 123123123UL };
	//
	uint32_t key[] = { 0x6b8b4567L, 0x327b23c6L, 0xf3aaa47L, 0x2c613bbfL };
	uint32_t nonce[2];

	_delay_ms(240);
	_delay_ms(240);
	_delay_ms(240);
	_delay_ms(240);

	wiegand_init();
	printf("LEDComm Reader Starting ....\n");
	
	STATUSDDR = 0xff;

	while(1)
	{
		char s = get_state();
		switch(s)
		{
			case STATE_WAITING:
				// wait for key to be present
				if(ledcomm_client_present()) {
					set_bit(STATUSPORT, STATUSBLUEPIN);
					unset_bit(STATUSPORT, STATUSREDPIN);
					unset_bit(STATUSPORT, STATUSGREENPIN);
					char tmp[LEN_CMD_NONCE];
					char buf[8];
					char *c, i;

					printf("Found client, sending data\n");


					// send challenge
					tmp[0] = CMD_NONCE;
					nonce[0] = rg_rand();
					nonce[1] = rg_rand();

					tea_encipher_32(nonce, (uint32_t *) buf, key);


					c = (char *) nonce;
					for(i = 0; i < 8; i++) {
						printf("%x\n", c[i]);
					}

					memcpy(tmp + 1, buf, sizeof(nonce));

					printf("sent nonce=%lx %lx (%d bytes)\n", nonce[0], nonce[1], sizeof(nonce));


					ledcomm_send_buffer(tmp, LEN_CMD_NONCE);

					// wait up to a quarter second for the response
					change_state(STATE_NONCE_SENT, 254);
					printf("Waiting for response...\n");
				} else {
					set_bit(STATUSPORT, STATUSREDPIN);
					unset_bit(STATUSPORT, STATUSBLUEPIN);
					unset_bit(STATUSPORT, STATUSGREENPIN);
				}
				break;
			case STATE_NONCE_SENT:
				// when we get the response, check the response
				if(ledcomm_rx_len() >= LEN_CMD_CRYPT) {
					volatile char *tmp = ledcomm_get_rxbuf();

					printf("Got data back\n");

					uint32_t newnonce[2];	
					uint32_t check_nonce[2];
					uint16_t client_id;

					if(tmp[0] != CMD_CRYPT) {
						printf("ERR: got back non crypt command (%d)\n", tmp[0]);
					}

					memcpy((char *) &client_id, (char *) tmp+1, 2);
					printf("client_id=%d\n", client_id);

					memcpy((char *) newnonce, (char *) tmp+3, sizeof(newnonce));

					tea_decipher_32(newnonce, check_nonce, key);

					if(check_nonce[0] == (nonce[0]+1) && check_nonce[1] == (nonce[1]+client_id)) {
						// bcr6: In this case, we're happy, everything is totally good
						printf("client %d ACCEPTED!!!\n", client_id);
						set_bit(STATUSPORT, STATUSGREENPIN);
						unset_bit(STATUSPORT, STATUSBLUEPIN);
						unset_bit(STATUSPORT, STATUSREDPIN);

						struct wiegand26 s;
						char buf[4];

						s.facility = 1;
						s.card = client_id;

						// if we're happy, send a wiegand message 
						// to the access control server to that effect.
						
						wiegand26_pack(&s, buf);
						wiegand_write(buf, 26);
						printf("done with wiegand\n");
					} else {
						printf("client not accepted\ncheck_nonce = %lx %lx\n",
							check_nonce[0], check_nonce[1]);
					}
					int III;
					for(III = 0; III < 70; III++) {
						_delay_ms(240);
					}

					ledcomm_rx_reset();
					rd_state = STATE_WAITING;
				}
		}
	}

}
