#include <stdio.h>
#include <stdlib.h>
#include <avr/io.h>
#include "util_macros.h"


/* This is a "one byte at a time" hash found at:
 *       http://www.ussg.iu.edu/hypermail/linux/kernel/9809.1/0741.html and
 *       http://burtleburtle.net/bob/hash/doobs.html
 *
 *       It performs reasonable hashing on most inputs, and is fast.  It also does not
 *       require any lookup tables (e.g., using CRC as a hash does).
 *       -- cgd3
 */
static uint32_t rg_one_at_a_time(unsigned char const *key, char len)

{
	uint32_t hash = 0x9e3779b9; /* Or any arbitrary value */

	char i;

	for (i=0; i < len; ++i) {
		hash += key[(int) i];
		hash += (hash << 10);
		hash ^= (hash >> 6);
	}
	hash += (hash << 3);
	hash ^= (hash >> 11);
	hash += (hash << 15);

	return hash;
}

void rg_init() {
	char bit_num;
	uint32_t rg_seed;

	

	bit_num = 0;

	// bcr6: seed the random number generator 
	// with noise off the ADC		
//	set_bit(ADMUX, ADLAR);		

	set_bit(ADCSRA, ADPS0);
	set_bit(ADCSRA, ADPS1);
	set_bit(ADCSRA, ADPS2);
	set_bit(ADCSRA, ADEN);

	set_bit(DDRA, 7);
	set_bit(ADMUX, MUX0);
	set_bit(ADMUX, MUX1);
	set_bit(ADMUX, MUX2);


	//set_bit(ADCSRA, ADSC);
	rg_seed = 0;
	while(bit_num < 32) {
		char low, high;

		set_bit(ADCSRA, ADSC);

		loop_until_bit_is_clear(ADCSRA, ADSC);

		low = ADCL;
		high = ADCH;

		if(bit_is_set(low, 0)) {
			set_bit(rg_seed, bit_num);
		}

		if(bit_is_set(low, 1)) {
			set_bit(rg_seed, bit_num + 1);
		}

		/*
		printf("%d: %d\n", bit_num, low & 0x3);
		for(i = 0; i < 10; i++) {
			_delay_ms(200);
		}
		*/

		//_delay_ms(200);
		bit_num += 2;
	}


	// bcr6: Disable the ADC
	unset_bit(ADCSRA, ADEN);
	rg_seed = rg_one_at_a_time((char *) &rg_seed, sizeof(rg_seed));
	srandom(rg_seed);
}

uint32_t rg_rand() {
	return random();
}

