#include "wiegand.h"
#include <avr/interrupt.h>
#include <stdio.h>

#ifndef WIEGAND_WRITE_PORT
//	#warning WIEGAND_WRITE_PORT not defined, write support disabled
#endif


volatile unsigned int _wiegand_count;
volatile unsigned int state;
volatile unsigned char _wiegand_bcount;
volatile unsigned char _wiegand_ready;
volatile unsigned char _wiegand_byte;
volatile char _wiegand_bit;
unsigned char *_wiegand_buf;
unsigned char _wiegand_maxlen;

static void _wiegand_register_bit(char bit);
static void _wiegand_finish_message();

// TODO: remove me!!!
void set_led_green() {}
void set_led_red() {}


void wiegand_timer_tick() {
	if(_wiegand_count == 1) {
		_wiegand_finish_message();
		_wiegand_ready = 1;
	}

	if(_wiegand_count) {
		_wiegand_count--;
	}	
}

void print_binary(char *s, uint16_t x) {
	char i;

	printf("%s: ", s);
	for(i = 15; i >= 0; i--) {
		if(get_bit(x, i)) {
			printf("1");
		} else {
			printf("0");
		}
	}
	printf("\n");
}

void print_binary8(uint8_t x) {
	char i;

	for(i = 7; i >= 0; i--) {
		if(get_bit(x, i)) {
			printf("1");
		} else {
			printf("0");
		}
	}
	printf("\n");
}

SIGNAL(INT0_vect) {
	_wiegand_register_bit(0);
}

SIGNAL(INT1_vect) {
	_wiegand_register_bit(1);
}

void _wiegand_register_bit(char bit) {
	if(bit_is_set(PIND, 2) && bit_is_set(PIND, 3)) {
		return;
	}

	if(!bit_is_set(PIND, 2) && !bit_is_set(PIND, 3)) {
		return;
	}

	if(_wiegand_bcount >= _wiegand_maxlen) {
		return;
	}

#if 0
	if(_wiegand_count >= 896) {
		return;
	}
#endif

	if(bit) {
		set_bit(_wiegand_buf[_wiegand_byte], _wiegand_bit);
	} else {
		unset_bit(_wiegand_buf[_wiegand_byte], _wiegand_bit);
	}

	if(_wiegand_bit == 0) {
		_wiegand_bit = 7;
		_wiegand_byte++;
	} else {
		_wiegand_bit--;
	}

	_wiegand_count = 900;
	_wiegand_bcount++;
	//PORTB = ~_wiegand_bcount;
	set_led_green();
}

void _wiegand_finish_message() {
	_wiegand_ready = 1;
}

void wiegand_init() {
	// bcr6: set it up such that we trigger on the falling edge
	// edge of both external interrupts
	unset_bit(MCUCR, ISC00);
	set_bit(MCUCR, ISC01);

	unset_bit(MCUCR, ISC10);
	set_bit(MCUCR, ISC11);

	_wiegand_bcount = 0;
	_wiegand_bit = 7;
	_wiegand_count = 0;
	_wiegand_ready = 0;
	_wiegand_byte = 0;

#ifdef WIEGAND_WRITE_PORT
	set_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN0);
	set_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN1);
#endif
}

unsigned char wiegand_read(unsigned char maxlen, char *buf) {
	wiegand_init();

	_wiegand_maxlen = maxlen;
	_wiegand_buf = buf;

	//printf("maximum allowed length=%d\n", maxlen);

	set_led_red();

	set_bit(GICR, INT1);
	set_bit(GICR, INT0);
	loop_until_bit_is_set(_wiegand_ready, 0);
	unset_bit(GICR, INT1);
	set_bit(GICR, INT0);

	set_led_red();
	return _wiegand_bcount;
}

char oparity(char byte) {
	char i;
	char rv = 0;

	for(i = 0; i < 7; i++) {
		if(byte & 1) {
			rv++;
		}
		
		byte >>= 1;
	}

	return rv % 2;
}

char wiegand26_take_eparity(char *mem) {
	char tmp;

	tmp = oparity(mem[0] & 0xfe); // mask out low
	tmp ^= oparity(mem[1] & 0x1f); // and then five bits of the next one

	return !tmp;
}

char wiegand26_take_oparity(char *mem) {
	char tmp;

	tmp = oparity(mem[2]);
	tmp ^= oparity(mem[1] & 0xe0);
	tmp ^= oparity(mem[3] & 0x01); // one low bit

	return tmp;
}


int wiegand26_unpack(char *mem, struct wiegand26 *out) 
{
	char eb,ob;

	eb = get_bit(mem[0], 7);
	ob = get_bit(mem[3], 6);

	printf("unpacking...\n");

	out->facility = mem[0] << 1;
	out->facility |= get_bit(mem[1], 7);

	uint16_t tmp;

	tmp = mem[1];
	print_binary("tmp(1)", tmp);
	out->card = tmp << 9;
	print_binary("card(1)", out->card);

	tmp = mem[2];
	tmp &= 0xff;
	print_binary("tmp(2)", tmp);

	out->card |= tmp << 1;
	print_binary("card(2)", out->card);
	out->card |= get_bit(mem[3], 7);
	print_binary("card(3)", out->card);
	
	if(wiegand26_take_oparity(mem) != ob) {
		return -1;	
	}

	if(wiegand26_take_eparity(mem) != eb) {
		return -1;	
	}

	return 0;
}

int wiegand30_unpack(char *mem, struct wiegand30 *out) {
	uint16_t tmp;

	out->facility = 0;
	out->facility = (mem[0] >> 3) & 0x0f;


	tmp = mem[1];
	out->card = 0;
	out->card |= (tmp << 13);

#if 0
	printf("tmp=");
	print_binary(tmp);


	printf("card(1)=");
	print_binary(out->card);

	printf("mem(2)=");
	print_binary8(mem[2]);
#endif

	tmp = mem[2];
	//printf("tmp=");
	//print_binary(tmp);

	tmp <<= 5;
	//printf("tmp(shifted)=");
	//print_binary(tmp);

	//out->card |= ((uint16_t) mem[2]) << 5;
	out->card |= (tmp & 0x1fff);

	//printf("card=\n");
	//print_binary(out->card);

	out->card |= ((((uint16_t) mem[3]) >> 3) & 31);


	//printf("card=%u\n", out->card);
	//print_binary(out->card);

	out->card = ~out->card;

	//printf("card(inv)=");
	print_binary("card", out->card);

	return 0;
}

void wiegand26_pack(struct wiegand26 *in, char *mem)
{
	char eb;
	char ob;


	mem[0] = (in->facility >> 1) & 0x7f;

	print_binary("card", in->card);

	uint16_t hb = (in->card >> 9) & 0x7f;
	print_binary("hb(1)", hb);
	mem[1] = (in->facility << 7) | hb;
	print_binary8(mem[1]);

	hb = (in->card >> 1) & 0xff;
	print_binary("hb(2)", hb);
	mem[2] = (in->card >> 1) & 0xff; 		
	print_binary8(mem[2]);

	mem[3] = get_bit(in->card, 0) << 7;
	print_binary8(mem[3]);

	eb = wiegand26_take_eparity(mem);
	ob = wiegand26_take_oparity(mem);

	mem[0] |= eb << 7;
	mem[3] |= ob << 6;
}


#ifdef WIEGAND_WRITE_PORT

static char _wiegand_cur_bit;
volatile static char _wiegand_state;
volatile static unsigned char _wiegand_tr;
#endif

enum {
	WS_SEND,
	WS_WAIT,
	WS_DONE
};


#define TICKS_SEND 13
#define TICKS_WAIT 500

#ifdef WIEGAND_WRITE_PORT

SIGNAL(TIMER2_COMP_vect) {				
	if(_wiegand_state == WS_WAIT) {
		// HACK: this assumes what remains fits in a byte
		if(_wiegand_tr > 0) {
			OCR2 = _wiegand_tr;			
			_wiegand_tr = 0;
			return;
		}

		char b = get_bit(_wiegand_buf[_wiegand_byte], _wiegand_bit);

		if(--_wiegand_bit < 0) {
			_wiegand_bit = 7;
			_wiegand_byte++;
		}

		_wiegand_bcount++;
		
		if(b) {
			unset_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN1);
		} else {
			unset_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN0);
		}

		OCR2 = TICKS_SEND;
		_wiegand_state = WS_SEND;

	} else if(_wiegand_state == WS_SEND) {
		set_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN0);
		set_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN1);

		_wiegand_tr = TICKS_WAIT - 0xff;
		if(_wiegand_bcount < _wiegand_maxlen) {
			OCR2 = 0xff;
			_wiegand_state = WS_WAIT;
		} else {
			unset_bit(TIMSK, OCIE2);
			_wiegand_state = WS_DONE;
		}
	}
//	DDRB = 0xff;
	//PORTB = ~_wiegand_bcount;
}


int wiegand_write(char *mem, int len) {
	_wiegand_cur_bit = 0;
	_wiegand_buf = mem;	
	_wiegand_maxlen = len;

	_wiegand_byte = 0;
	_wiegand_bit = 7;

	printf("sbyte=%d, sbit=%d\n", _wiegand_byte, _wiegand_bit);

	_wiegand_state = WS_WAIT;
	_wiegand_tr = 0;


	OCR2 = 1; // fire right away

	set_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN0);
	set_bit(WIEGAND_WRITE_PORT, WIEGAND_WRITE_PIN1);

	set_bit(TIMSK, OCIE2);
	TCCR2 = 3; // prescalar/64
	set_bit(TCCR2, WGM21);
	unset_bit(TCCR2, WGM20);

	while(_wiegand_state != WS_DONE) {
		// do nothing	
	}

	wiegand_init();

	return 0;
}
#endif
